use std::{
    collections::{hash_map::Entry, HashMap},
    sync::Arc,
};

use parking_lot::RwLock;
use protocol::*;
use qwer::{phashmap, phashset, PropertyHashMap, PropertyHashSet};

use crate::data::{self, ConfigAction, ConfigEventType, ConfigValue};

pub struct HollowGridManager {
    player: Arc<RwLock<PlayerInfo>>,
    map: RwLock<Option<HollowGridMapProtocolInfo>>,
    events: RwLock<HashMap<u64, EventInfo>>,
}

impl HollowGridManager {
    pub fn new(player: Arc<RwLock<PlayerInfo>>) -> Self {
        Self {
            player,
            map: RwLock::new(None),
            events: RwLock::new(HashMap::new()),
        }
    }

    pub fn get_cur_position_in_hollow(&self) -> u16 {
        self.map.read().as_ref().unwrap().start_grid
    }
    
    pub fn get_cur_chessboard_id(&self) -> i32 {
        self.map.read().as_ref().unwrap().chessboard_id
    }

    pub fn get_col_num(&self) -> i32 {
        <u8 as Into<i32>>::into(self.map.read().as_ref().unwrap().col).clone()
    }

    pub fn move_to(
        &self,
        destination_grid: u16,
        scene_uid: u64,
    ) -> (PtcHollowGridArg, Option<PtcSyncHollowEventInfoArg>) {
        let mut map = self.map.write();
        let map = map.as_mut().unwrap();
    
        let cur_hollow = map.chessboard_id.clone();

        map.start_grid = destination_grid;
        let grid = map.grids.get_mut(&destination_grid).unwrap();

        self.update_position_to_scene(scene_uid, destination_grid);
        
        let entry_uid = (scene_uid * 1000) + (destination_grid as u64);//Adds new entries infinitely, works for now, but should                                                                          //be replaced with a proper clear_event_entries function
        let mut events = self.events.write();
        let sync_event_info =
            if let Entry::Vacant(entry) = events.entry(u64::from(entry_uid)) {
                let event_info = EventInfo {
                    id: 1000,
                    cur_action_id: 1001,
                    action_move_path: vec![1001],
                    state: EventState::WaitingClient,
                    prev_state: EventState::Running,
                    cur_action_info: ActionInfo::None {},
                    cur_action_state: ActionState::Init,
                    predicated_failed_actions: phashset![],
                    stack_frames: Vec::new(),
                };

                entry.insert(event_info.clone());

                Some(PtcSyncHollowEventInfoArg {
                    event_graph_uid: u64::from(destination_grid),
                    hollow_event_template_id: grid.grid.event_graph_info.hollow_event_template_id,
                    event_graph_id: grid.grid.event_graph_info.hollow_event_template_id,
                    updated_event: event_info,
                    specials: phashmap![],
                })
            } else {
                None
            };

        if !grid.grid.event_graph_info.finished {
            grid.grid.flag |= HollowGridFlag::Travelled as i32;
            grid.grid.flag |= HollowGridFlag::ShowEventID as i32;
            grid.grid.flag &= !(HollowGridFlag::Guide as i32);
            grid.grid.flag &= !(HollowGridFlag::CanTriggerEvent as i32);
            grid.grid.flag &= !(HollowGridFlag::ShowEventType as i32);
            grid.grid.event_graph_info.finished = true;

            grid.grid.event_graph_info.fired_count = 2;
        }

        let mut update_grid_on_move = PtcHollowGridArg {
            player_uid: self.player.read().uid.unwrap(),
            is_partial: true,
            scene_uid,
            hollow_level: 1,
            grids: HashMap::from([(destination_grid, grid.clone()),]),
        };
        
        if cur_hollow == 199003706 {
 //Custom Event 1: for finding the carrot card (rescue)
            if let Some(grid) = map.grids.get_mut(&(destination_grid + 17)) {
                if grid.grid.node_state == NodeState::Target {
                grid.grid.flag &= !(HollowGridFlag::Visible as i32);

                    update_grid_on_move.grids.insert(destination_grid + 17, grid.clone());
                

                    if let Some(grid) = map.grids.get_mut(&12) {
                        grid.grid.flag |= HollowGridFlag::Visible as i32;

                            update_grid_on_move.grids.insert(12, grid.clone());
                    }
                }
            };
 //Custom Event 2: back to start (rescue)
            if let Some(grid) = map.grids.get_mut(&(destination_grid + 1)) {
                if grid.grid.node_state == NodeState::Target {
                    grid.grid.flag |= HollowGridFlag::Visible as i32
                                   | HollowGridFlag::CanMove as i32
                                   | HollowGridFlag::ShowEventType as i32;

                    update_grid_on_move.grids.insert(destination_grid + 1, grid.clone());
                }
            };
 //Custom Event 3: ambush (rescue)
            if destination_grid == 65 {
                if let Some(grid) = map.grids.get_mut(&(destination_grid + 1)) {
                        grid.grid.flag |= HollowGridFlag::Visible as i32
                                       | HollowGridFlag::CanMove as i32
                                       | HollowGridFlag::ShowEventType as i32;

                        update_grid_on_move.grids.insert(destination_grid + 1, grid.clone());
                }
                if let Some(grid) = map.grids.get_mut(&(destination_grid - 1)) {
                        grid.grid.flag |= HollowGridFlag::Visible as i32
                                       | HollowGridFlag::CanMove as i32
                                       | HollowGridFlag::ShowEventType as i32;

                        update_grid_on_move.grids.insert(destination_grid - 1, grid.clone());
                }
            };
        }
        if cur_hollow == 101010902 {

 //Custom Event 4: show friendly units (time trial)
            if destination_grid == 38 {
                let friendly_tiles: [u16; 2] = [31, 53];
                for x in friendly_tiles {
                    if let Some(grid) = map.grids.get_mut(&x) {
                        if !grid.grid.event_graph_info.finished {
                                grid.grid.flag |= HollowGridFlag::Visible as i32
                                               | HollowGridFlag::CanMove as i32
                                               | HollowGridFlag::ShowEventType as i32;

                                update_grid_on_move.grids.insert(x, grid.clone());
                        };
                    };
                }
            }
 //Custom Event 5: heart (time trial)
            if destination_grid == 40 {
                let red_tiles: [u16; 11] = [13, 15, 23, 24, 25, 26, 27, 35, 36, 37, 47];
                for x in red_tiles {
                    if let Some(grid) = map.grids.get_mut(&x) {
                            grid.grid.flag |= HollowGridFlag::Visible as i32
                                           | HollowGridFlag::CanMove as i32
                                           | HollowGridFlag::ShowEventType as i32;

                            update_grid_on_move.grids.insert(x, grid.clone());
                    };
                };
            }
 //Custom Event 6: move friendly units (time trial)
            if destination_grid == 39 {
                let friendly_tiles: [u16; 2] = [31, 53];
                let friendly_tiles_move: [u16; 2] = [30, 52];
                for x in friendly_tiles {
                    if let Some(grid) = map.grids.get_mut(&x) {
                            grid.grid.flag &= !(HollowGridFlag::Visible as i32);
                            grid.grid.event_graph_info.finished = true;

                            update_grid_on_move.grids.insert(x, grid.clone());
                    }
                };
                for x in friendly_tiles_move {
                    if let Some(grid) = map.grids.get_mut(&x) {
                            grid.grid.flag |= HollowGridFlag::Visible as i32
                                           | HollowGridFlag::CanMove as i32
                                           | HollowGridFlag::ShowEventType as i32;

                            update_grid_on_move.grids.insert(x, grid.clone());
                    }
                };
            }
        }
        if cur_hollow == 199003501 {

 //Custom Event 7: no way back (descent)
            let triggers: [u16; 2] = [14, 16];
            for x in triggers {
                if destination_grid == x {
                    let destroyed_path: [u16; 2] = [5, 15];
                    for x in destroyed_path {
                        if let Some(grid) = map.grids.get_mut(&x) {
                                grid.grid.flag &= !(HollowGridFlag::Visible as i32);

                                update_grid_on_move.grids.insert(x, grid.clone());
                        };
                    };
                }
            }

 //Custom Event 8: barriers (descent)
            let triggers: [u16; 2] = [43, 47];
            for x in triggers {
                if destination_grid == x {
                    let barriers: [u16; 2] = [42, 48];
                    for x in barriers {
                        if let Some(grid) = map.grids.get_mut(&x) {
                                grid.grid.flag |= HollowGridFlag::ShowEventID as i32;
                                grid.grid.event_graph_info.hollow_event_template_id = 2020825;
                                grid.grid.event_graph_info.finished = false;

                                update_grid_on_move.grids.insert(x, grid.clone());
                        };
                    };  
                }
            }

 //Custom Event 9: no way back 2 (descent)
            let triggers: [u16; 2] = [54, 56];
            for x in triggers {
                if destination_grid == x {
                    if let Some(grid) = map.grids.get_mut(&45) {
                            grid.grid.flag &= !(HollowGridFlag::Visible as i32);
                            grid.grid.node_visible = NodeVisible::All;

                            update_grid_on_move.grids.insert(45, grid.clone());
                    };
                }
            }

 //Custom Event 10: barriers 2 (descent)
            let triggers: [u16; 2] = [94, 96];
            for x in triggers {
                if destination_grid == x {
                    let barriers: [u16; 2] = [93, 97];
                    for x in barriers {
                        if let Some(grid) = map.grids.get_mut(&x) {
                                grid.grid.flag |= HollowGridFlag::ShowEventID as i32;
                                grid.grid.event_graph_info.hollow_event_template_id = 2020825;
                                grid.grid.event_graph_info.finished = false;

                                update_grid_on_move.grids.insert(x, grid.clone());
                        };
                    };  
                }
            }

 //Custom Event 11: no way back 3 (descent)
            let triggers: [u16; 2] = [104, 106];
            for x in triggers {
                if destination_grid == x {
                    if let Some(grid) = map.grids.get_mut(&95) {
                            grid.grid.flag &= !(HollowGridFlag::Visible as i32);
                            grid.grid.node_visible = NodeVisible::All;

                            update_grid_on_move.grids.insert(95, grid.clone());
                    };
                }
            }

 //Custom Event 12: barriers 3 (descent)
            let triggers: [u16; 2] = [124, 126];
            for x in triggers {
                if destination_grid == x {
                    let barriers: [u16; 2] = [114, 116];
                    for x in barriers {
                        if let Some(grid) = map.grids.get_mut(&x) {
                                grid.grid.flag |= HollowGridFlag::ShowEventID as i32;
                                grid.grid.event_graph_info.hollow_event_template_id = 2020825;
                                grid.grid.event_graph_info.finished = false;

                                update_grid_on_move.grids.insert(x, grid.clone());
                        };
                    };  
                }
            }
        }
        if cur_hollow == 199003705 {

 //Custom Event 13: door (Lost and Found)
            let blocked_tiles: [u16; 2] = [3, 5];
            if destination_grid == 42 {
                for x in blocked_tiles {
                    if let Some(grid) = map.grids.get_mut(&x) {
                            grid.grid.event_graph_info.hollow_event_template_id = 1000;


                            update_grid_on_move.grids.insert(x, grid.clone());
                    };
                }
                if let Some(grid) = map.grids.get_mut(&4) {
                        grid.grid.node_state = NodeState::Unlocked;
                        grid.grid.flag |= HollowGridFlag::Visible as i32
                                       | HollowGridFlag::CanMove as i32
                                       | HollowGridFlag::ShowEventType as i32;
                        grid.grid.event_graph_info.hollow_event_template_id = 2020826;

                        update_grid_on_move.grids.insert(4, grid.clone());
                };
            }
            for x in blocked_tiles {
                if destination_grid == x { 
                    if let Some(grid) = map.grids.get_mut(&4) {
                        if grid.grid.node_state == NodeState::Unlocked {
                                grid.grid.event_graph_info.finished = true;
                                grid.grid.event_graph_info.hollow_event_template_id = 1000;

                                update_grid_on_move.grids.insert(4, grid.clone());
                        };
                    }; 
                }
            } 
        }
        (

            update_grid_on_move,
            sync_event_info,

        )
    }

    //pub fn clear_event_entries(
        //&self,
        //scene_uid: u64,
    //) -> (PtcHollowGridArg, Option<PtcSyncHollowEventInfoArg>) {
        //let mut map = self.map.write();
        //let map = map.as_mut().unwrap();

       //}
    //}

    pub fn check_around(
        &self,
        destination_grid: u16,
        scene_uid: u64,
    ) -> PtcHollowGridArg {
        let mut map = self.map.write();
        let map = map.as_mut().unwrap();
        
        let mut grid_update_around = PtcHollowGridArg {
            player_uid: self.player.read().uid.unwrap(),
            is_partial: true,
            scene_uid,
            hollow_level: 1,
            grids: HashMap::new(),
        };

        map.start_grid = destination_grid;

        let num_of_col = (map.col as u16).clone();

        let mut check = 0;

        let mut uid_around: u16;

        let uid_around_right = (destination_grid + 1) as u16;
        let uid_around_left = (destination_grid - 1) as u16;
        let uid_around_down = (destination_grid + num_of_col) as u16;
        let uid_around_up = (destination_grid - num_of_col) as u16;

        //Check if you are at the edge on the sides of a level
        if destination_grid % num_of_col == 0 {
            uid_around = uid_around_right;
            check = 2;
        } else if (destination_grid + 1) % num_of_col == 0 {
             uid_around = uid_around_left;
             check = 2;
        } else {
             uid_around = uid_around_right;
        };

        //Checking for tiles with VisibleAtGridAround parameter around the player
        while check < 5 {
            if check == 1 {
                uid_around = uid_around_left;
                check = 2;
            }
            else if check == 3 {
                uid_around = uid_around_down;
            }
            else if check == 4{
                uid_around = uid_around_up;
            };

            if let Some(grid) = map.grids.get_mut(&uid_around) {
                if grid.grid.node_visible == NodeVisible::VisibleAtGridAround {
                grid.grid.flag |= HollowGridFlag::Visible as i32
                               | HollowGridFlag::CanMove as i32
                               | HollowGridFlag::ShowEventType as i32;

                    grid_update_around.grids.insert(uid_around, grid.clone());
                }
            };
            check += 1;
        };
                grid_update_around
    }

    pub fn battle_finished(&self) -> (PtcSyncHollowEventInfoArg, bool) {
        let map = self.map.read();
        let map = map.as_ref().unwrap();
        let cur_grid = map.grids.get(&map.start_grid).unwrap();

        let graph =
            data::get_event_graph(cur_grid.grid.event_graph_info.hollow_event_template_id).unwrap();

        let mut hollow_finished = false;
        if let Some(event) = graph.events.get(&ConfigEventType::OnEnd) {
            if let Some(ConfigAction::ConfigFinishHollow) = event.actions.first() {
                hollow_finished = true;
            }
        }

        (
            PtcSyncHollowEventInfoArg {
                event_graph_uid: u64::from(map.start_grid),
                hollow_event_template_id: cur_grid.grid.event_graph_info.hollow_event_template_id,
                event_graph_id: cur_grid.grid.event_graph_info.hollow_event_template_id,
                updated_event: EventInfo {
                    id: 1000,
                    cur_action_id: 2001,
                    action_move_path: vec![1001, 1002, 2001],
                    state: EventState::WaitingClient,
                    prev_state: EventState::Running,
                    cur_action_info: ActionInfo::None {},
                    cur_action_state: ActionState::Init,
                    predicated_failed_actions: phashset![],
                    stack_frames: Vec::new(),
                },
                specials: phashmap![],
            },
            hollow_finished,
        )
    }

    pub fn get_cur_event_template_id(&self) -> i32 {
        let map = self.map.read();
        let map = map.as_ref().unwrap();
        let cur_grid = map.grids.get(&map.start_grid).unwrap();

        cur_grid.grid.event_graph_info.hollow_event_template_id
    }

    fn update_position_to_scene(&self, scene_uid: u64, pos: u16) {
        let mut player = self.player.write();
        let scene = player
            .dungeon_collection
            .as_mut()
            .unwrap()
            .scenes
            .as_mut()
            .unwrap()
            .get_mut(&scene_uid)
            .unwrap();

        if let SceneInfo::Hollow {
            sections_info,
            hollow_event_graph_uid,
            ..
        } = scene
        {
            let section = sections_info.get_mut(&1).unwrap();
            section.prev_grid_index = section.cur_grid_index;
            section.pos_before_move = section.cur_grid_index;
            section.cur_grid_index = pos;
            *hollow_event_graph_uid = u64::from(pos);
        } else {
            panic!("Unexpected scene type")
        }
    }

    pub fn run_event_graph(
        &self,
        event_graph_uid: u64,
        _event_id: i32,
        move_path: Vec<i32>,
    ) -> (
        PtcSyncHollowEventInfoArg,
        PtcHollowGridArg,
        Option<i32>,
        bool,
    ) {
        let (player_uid, scene_uid) = {
            let player = self.player.read();

            (player.uid.unwrap(), player.scene_uid.unwrap())
        };

        let mut map = self.map.write();
        let map = map.as_mut().unwrap();

        let mut trigger_battle_id = None;
        let mut grid_update = PtcHollowGridArg {
            player_uid,
            is_partial: true,
            scene_uid,
            hollow_level: 1,
            grids: HashMap::new(),
        };

        let mut hollow_finished = false;

        let sync_hollow_event = {
            let info = map.grids.get(&(event_graph_uid as u16)).unwrap().clone();
            let graph =
                data::get_event_graph(info.grid.event_graph_info.hollow_event_template_id).unwrap();

            let cur_hollow = map.chessboard_id.clone();
            
            let num_of_col = (map.col as i32).clone();

            let mut last_action = &ConfigAction::ConfigEmpty;

            for id in &move_path {
                let index = (id % 1000) - 1;
                let event = if id / 1000 == 1 {
                    graph.events.get(&ConfigEventType::OnStart)
                } else {
                    graph.events.get(&ConfigEventType::OnEnd)
                };
                if let Some(action) = event.unwrap().actions.get(index as usize) {
                    last_action = action;

                    match action {
                        ConfigAction::ConfigSetMapState { x, y, .. } => {
                            let (ConfigValue::Constant(x), ConfigValue::Constant(y)) = (x, y)
                            else {
                                panic!("ConfigSetMapState: only constant values are supported");
                            };
                            
                            let uid = ((y * num_of_col) + x) as u16;
                            if let Some(info) = map.grids.get_mut(&uid) {
                                info.grid.flag |= HollowGridFlag::Visible as i32
                                    | HollowGridFlag::CanMove as i32
                                    | HollowGridFlag::ShowEventType as i32;

                                grid_update.grids.insert(uid, info.clone());
                            }
                        }
                        ConfigAction::ConfigTriggerBattle { .. } => {

                        //Battle lists. Tied to ChessboardID
                            if cur_hollow == 199003706 { //Rescue
                            trigger_battle_id = 
                                Some(match info.grid.event_graph_info.hollow_event_template_id {
                                    1000107 => 10102005, //level end
                                    1000208 => 10102002, //normal, ally 1
                                    1000209 => 10301232, //normal
                                    1000210 => 30141011, //normal, time trial
                                    1000310 => 30122022, //normal
                                    1000311 => 30112005, //normal
                                    1000312 => 10102001, //normal, ally 2
                                    1000313 => 10301234, //elite
                                    1000317 => 30123028, //elite
                                    1000318 => 10102016, //elite
                                    1000319 => 10102003, //normal, unused
                                    1010609 => 30123047, //elite
                                    1010707 => 10102017, //10301234, normal
                                    1010708 => 30122009, //normal
                                    2020844 => 10301216, //dire wolf
                                    2000102 => 30112010, //normal
                                    1009 => 10102004, //normal animation, coded as elite, unused
                                    2012 => 10301226, //boss
                                    _ =>30121012, //failsafe for other battle events
                                })
                            } else if cur_hollow == 199003501{ //High-Risk 1
                                trigger_battle_id = 
                                Some(match info.grid.event_graph_info.hollow_event_template_id {
                                    1000107 => 10301209, //level end
                                    1000313 => 30123049, //elite
                                    1000317 => 30123044, //elite
                                    1000318 => 30122016, //elite
                                    1000319 => 30122026, //normal
                                    1000209 => 30141021, //normal
                                    1000208 => 30112015, //normal
                                    1000310 => 30123011, //normal
                                    1010609 => 10301238, //elite
                                    1010707 => 30122003, //normal
                                    2000102 => 30122036, //normal
                                    2012 => 10301202, //boss
                                    _ =>30121012, //failsafe for other battle events
                                })
                            } else if cur_hollow == 199003705{ //Lost and found
                                trigger_battle_id = 
                                Some(match info.grid.event_graph_info.hollow_event_template_id {
                                    1000107 => 10301241, //level end
                                    1000313 => 30132031, //elite
                                    1000317 => 30123045, //elite
                                    1000318 => 30133008, //elite
                                    1000319 => 30132021, //normal
                                    1000208 => 30112008, //normal
                                    1000209 => 30123004, //normal
                                    1000310 => 30122030, //normal
                                    1010609 => 30132014, //elite
                                    1010707 => 30122013, //normal
                                    2000102 => 10101027, //normal
                                    2012 => 30133018, //10301209, boss
                                    _ =>30121012, //failsafe for other battle events
                                })
                            } else if cur_hollow == 2020701{ //Prologue (Where Despair Breeds)
                                trigger_battle_id = 
                                Some(match info.grid.event_graph_info.hollow_event_template_id {
                                    1000107 => 10101002, //level end
                                    1000208 => 10101008, //normal
                                    1000209 => 10101006, //normal
                                    1000310 => 10101003, //normal
                                    2012 => 30123046, //danger
                                    _ => 10101001, //failsafe for other battle events
                                })
                            } else {
                                trigger_battle_id = 
                                Some(match info.grid.event_graph_info.hollow_event_template_id {
                                    _ => 19900029, //normal, time trial
                                })
                            };
                        }
                        
 
                        ConfigAction::ConfigFinishHollow => {
                            hollow_finished = true;
                        }
                        //unfinished teleport event 
                        //ConfigAction::ConfigTransfer{x, y, ..} => {
                            //let (ConfigValue::Constant(x), ConfigValue::Constant(y)) = (x, y)
                            //else {
                                //panic!("ConfigTransfer: only constant values are supported");
                            //};
                        _ => {}
                    };
                }
            }

            let mut action_move_path = move_path;

            let last_client_action = *action_move_path.last().unwrap();
            let actions = if last_client_action / 1000 == 1 {
                &graph.events.get(&ConfigEventType::OnStart).unwrap().actions
            } else {
                &graph.events.get(&ConfigEventType::OnEnd).unwrap().actions
            };
            let state = if last_client_action == -1 {
                EventState::Finished
            } else if last_client_action % 1000 >= actions.len() as i32 {
                action_move_path.push(-1);
                EventState::Finished
            } else {
                if !matches!(last_action, ConfigAction::ConfigEmpty) {
                    action_move_path.push(last_client_action + 1);
                }

                EventState::WaitingClient
            };

            let finish_event = if let ConfigAction::ConfigTriggerBattle { .. } = last_action {
                PtcSyncHollowEventInfoArg {
                    event_graph_uid,
                    hollow_event_template_id: info.grid.event_graph_info.hollow_event_template_id,
                    event_graph_id: info.grid.event_graph_info.hollow_event_template_id,
                    updated_event: EventInfo {
                        id: 0,
                        cur_action_id: 0,
                        action_move_path: vec![],
                        state: EventState::Initing,
                        prev_state: EventState::Initing,
                        cur_action_info: ActionInfo::None {},
                        cur_action_state: ActionState::Init,
                        predicated_failed_actions: phashset![],
                        stack_frames: Vec::new(),
                    },
                    specials: phashmap![],
                }
            } else {
                PtcSyncHollowEventInfoArg {
                    event_graph_uid,
                    hollow_event_template_id: info.grid.event_graph_info.hollow_event_template_id,
                    event_graph_id: info.grid.event_graph_info.hollow_event_template_id,
                    updated_event: EventInfo {
                        id: 1000,
                        cur_action_id: *action_move_path.last().unwrap(),
                        action_move_path,
                        state,
                        prev_state: EventState::Running,
                        cur_action_info: ActionInfo::None {},
                        cur_action_state: ActionState::Init,
                        predicated_failed_actions: phashset![],
                        stack_frames: Vec::new(),
                    },
                    specials: phashmap![],
                }
            };

            tracing::info!("sending evt info: {:#?}", &finish_event);
            finish_event
        };

        (
            sync_hollow_event,
            grid_update,
            trigger_battle_id,
            hollow_finished,
        )
    }

    pub fn sync_hollow_maps(&self, player_uid: u64, scene_uid: u64) -> PtcSyncHollowGridMapsArg {
        PtcSyncHollowGridMapsArg {
            player_uid,
            scene_uid,
            hollow_level: 1,
            main_map: self.map.read().clone().unwrap(),
            time_period: TimePeriodType::Random, //Tied to ChessboardID, can be changed with PtcSyncSceneTimeArg in                                                                //world.rs for some IDs
            weather: WeatherType::Random, //Tied to ChessboardID
        }
    }

    pub fn init_prologue(&self) {
        *self.map.write() = Some(HollowGridMapProtocolInfo {
            row: 5,
            col: 11,
            start_grid: 22,
            grids: phashmap![
                (
                    48,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 12,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    7,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 10,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1017,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    24,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2682,
                            link_to: 12,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000103,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    28,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 12,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    3,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2682,
                            link_to: 6,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 2012,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::BattleElite,
                        use_perform: false,
                    }
                ),
                (
                    14,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2658,
                            link_to: 3,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::Dialog,
                        use_perform: false,
                    }
                ),
                (
                    4,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 12,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    36,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 3,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    29,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 13,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000310,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    49,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 9,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1018,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    27,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 7,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    6,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 12,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    16,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 3,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    22,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2686,
                            link_to: 4,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000101,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::Begin,
                        use_perform: false,
                    }
                ),
                (
                    30,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 12,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    18,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 3,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    38,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 3,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000208,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::BattleNormal,
                        use_perform: false,
                    }
                ),
                (
                    32,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 8,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000107,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::ChangeLevelInteract,
                        use_perform: false,
                    }
                ),
                (
                    47,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 5,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000209,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::BattleNormal,
                        use_perform: false,
                    }
                ),
                (
                    25,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2682,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000102,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::Dialog,
                        use_perform: false,
                    }
                ),
                (
                    5,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 14,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    31,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 35434,
                            link_to: 14,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000106,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::Dialog,
                        use_perform: false,
                    }
                ),
                (
                    23,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 35434,
                            link_to: 12,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000109,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::Dialog,
                        use_perform: false,
                    }
                )
            ],
            chessboard_id: 2020701,
        });
    }

  //Lost and Found
    pub fn init_lost(&self) {
        *self.map.write() = Some(HollowGridMapProtocolInfo {
            row: 15,
            col: 11,
            start_grid: 8,
            grids: phashmap![
                (
                    8,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2686,
                            link_to: 12,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000101,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::Begin,
                        use_perform: false,
                    }
                ),
                (
                    9,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 12,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    10,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 10,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    21,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 3,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    32,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 3,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    43,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 9,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    53,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 9,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    7,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2658,
                            link_to: 14,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    6,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2682,
                            link_to: 12,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 703,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    5,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 12,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 2020825,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    4,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 52024,
                            link_to: 12,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 2020825,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::Door,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    3,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 12,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 2020825,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    2,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 6,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    1,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 6,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 2010206,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    13,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 3,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    18,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2682,
                            link_to: 3,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000102,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::Dialog,
                        use_perform: false,
                    }
                ),
                (
                    29,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 13,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    28,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 12,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000311,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::BattleNormal,
                        use_perform: false,
                    }
                ),
                (
                    27,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::Dialog,
                        use_perform: false,
                    }
                ),
                (
                    50,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 8,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1017,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    17,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2682,
                            link_to: 0,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 2010206,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    137,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 0,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 2010206,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    139,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 0,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 2010206,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    149,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 0,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 2010206,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    64,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 0,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 2010206,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    76,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 0,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 2010206,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    118,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 0,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 2010206,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    98,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 0,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 2010206,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    111,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2682,
                            link_to: 0,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 2010206,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    35,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2682,
                            link_to: 0,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 2010206,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    83,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2682,
                            link_to: 0,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 2010206,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    115,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2682,
                            link_to: 0,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 2010206,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    81,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2682,
                            link_to: 0,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 2010206,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    26,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 14,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    30,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    41,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000317,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::BattleNormal,
                        use_perform: false,
                    }
                ),
                (
                    42,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1010612,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    71,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 3,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    82,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 35434,
                            link_to: 3,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1011110,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    102,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 12,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    93,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2658,
                            link_to: 3,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    104,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2658,
                            link_to: 13,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    116,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 3,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1016,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    127,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 3,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1016,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    138,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 1,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000107,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    105,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2682,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1018,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    79,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2658,
                            link_to: 10,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::BattleNormal,
                        use_perform: false,
                    }
                ),
                (
                    88,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2658,
                            link_to: 3,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    37,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    52,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 7,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    63,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 3,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    74,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 5,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1010609,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::BattleNormal,
                        use_perform: false,
                    }
                ),
                (
                    75,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 10,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    86,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 3,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    97,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 9,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000310,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    96,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 6,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1011105,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    106,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 12,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    36,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 7,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    49,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1010704,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    1,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 7,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    16,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 3,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    33,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 0,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 2010206,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    22,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 0,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    38,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 11,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000313,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::BattleNormal,
                        use_perform: false,
                    }
                ),
                (
                    25,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000208,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    24,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 13,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    23,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 6,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::ChangeLevelInteract,
                        use_perform: false,
                    }
                ),
                (
                    34,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 3,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::ChangeLevelInteract,
                        use_perform: false,
                    }
                ),
                (
                    45,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 7,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::ChangeLevelInteract,
                        use_perform: false,
                    }
                ),
                (
                    67,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 3,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::ChangeLevelInteract,
                        use_perform: false,
                    }
                ),
                (
                    56,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 3,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1010707,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::ChangeLevelInteract,
                        use_perform: false,
                    }
                ),
                (
                    114,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 3,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1011115,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::ChangeLevelInteract,
                        use_perform: false,
                    }
                ),
                (
                    46,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 52024,
                            link_to: 12,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1025,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    47,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 13,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000209,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::BattleNormal,
                        use_perform: false,
                    }
                ),
                (
                    60,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2682,
                            link_to: 3,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1017,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    31,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1010703,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::Dialog,
                        use_perform: false,
                    }
                ),
                (
                    48,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    77,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2682,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000319,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::BattleElite,
                        use_perform: false,
                    }
                ),
                (
                    103,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2682,
                            link_to: 14,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 2012,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::BattleElite,
                        use_perform: false,
                    }
                ),
                (
                    101,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2682,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000318,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::BattleNormal,
                        use_perform: false,
                    }
                ),
                (
                    107,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2682,
                            link_to: 9,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000218,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::Dialog,
                        use_perform: false,
                    }
                ),
                (
                    99,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2682,
                            link_to: 7,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 703,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::Dialog,
                        use_perform: false,
                    }
                ),
                (
                    78,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2658,
                            link_to: 13,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    90,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2658,
                            link_to: 3,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    100,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 35434,
                            link_to: 12,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000215,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    90,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2658,
                            link_to: 3,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    110,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 3,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    121,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 5,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    122,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 12,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    123,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 12,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    124,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 12,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    125,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 9,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                )
            ],
            chessboard_id: 199003705,
        });
    }
//Rescue mission:
    pub fn init_rescue(&self) {
        *self.map.write() = Some(HollowGridMapProtocolInfo {
            row: 11,
            col: 17,
            start_grid: 106,
            grids: phashmap![
                (
                    106,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2686,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000306,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::Begin,
                        use_perform: false,
                    }
                ),
                (
                    141,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2658,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    142,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2658,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    123,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2658,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    124,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2658,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    125,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2658,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    104,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    105,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::Target,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    107,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2658,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    108,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2658,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    109,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2682,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000210,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    90,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2658,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    73,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2682,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1010708,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    56,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2658,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    39,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2658,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    40,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2658,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    23,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 35434,
                            link_to: 10,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 2020825,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    24,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2682,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 2020825,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    6,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2682,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 2020825,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    7,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2682,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 2020825,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    26,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1022,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::Target,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    8,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 19115,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    25,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 19115,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    9,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 35434,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000208,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    10,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    11,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    28,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    45,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    62,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2658,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    63,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 35434,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1010714,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    61,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2658,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    38,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 52000,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    22,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 52000,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    21,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 52000,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    20,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 2020844,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    19,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    36,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    51,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000209,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    52,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    53,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    54,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    55,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    68,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 35434,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000205,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    69,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000310,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    70,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    85,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1010707,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    86,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    87,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    143,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2682,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1013,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    144,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2682,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1013,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    145,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2682,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 2000102,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    110,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1016,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    92,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1016,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    93,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1025,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    94,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1016,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    76,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1016,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    77,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1025,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    59,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1025,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    60,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1016,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    161,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2682,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000311,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    162,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2682,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1013,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    163,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2682,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000317,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    164,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2682,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000318,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    165,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2682,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1013,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    179,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2682,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1010703,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    180,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2682,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1010609,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    181,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2682,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1013,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    147,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2682,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1013,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    148,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2682,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000313,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    131,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2682,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1013,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    114,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2682,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1013,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    97,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2682,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 2012,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    79,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2658,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    80,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2658,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    12,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 19115,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 100000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    13,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 265016,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    14,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 265016,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000206,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    15,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 265016,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1025,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    16,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 10,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000312,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    33,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 265016,
                            link_to: 3,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    50,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 265016,
                            link_to: 3,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1025,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    67,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 265016,
                            link_to: 3,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    84,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 265016,
                            link_to: 9,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    83,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 265016,
                            link_to: 12,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    82,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 265016,
                            link_to: 5,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    65,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 265016,
                            link_to: 2,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000107,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::BattleEnd,
                        use_perform: false,
                    }
                ),
                (
                    64,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 2,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 2011311,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    66,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 2,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 2011311,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    101,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 2,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 2010907,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    99,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 2,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 2010907,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    48,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 2010907,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                )
            ],
            chessboard_id: 199003706,
        });
     }
//Time Trial
    pub fn init_time_trial(&self) {
        *self.map.write() = Some(HollowGridMapProtocolInfo {
            row: 11,
            col: 11,
            start_grid: 113,
            grids: phashmap![
                (
                    113,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2686,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000101,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::Begin,
                        use_perform: false,
                    }
                ),
                (
                    102,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2658,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    91,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2658,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    80,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2658,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    69,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2682,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000208,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    58,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 19115,
                            link_to: 6,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    59,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 19115,
                            link_to: 9,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    48,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 19115,
                            link_to: 6,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    49,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 19115,
                            link_to: 9,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    38,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 6,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1017,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    39,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    40,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000109,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    41,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1002,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    13,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 2020825,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    15,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 2020825,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    23,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 2020825,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    24,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 2020825,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    25,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 2020825,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    26,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 2020825,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    27,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 2020825,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    35,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 2020825,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    36,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 2020825,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    37,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 2020825,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    47,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 2020825,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    22,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    31,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 2010907,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    53,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 2010907,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    30,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 2010907,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    52,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 2010907,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                )
            ],
            chessboard_id: 101010902,
        });
     }
//Descent:
    pub fn init_descent(&self) {
        *self.map.write() = Some(HollowGridMapProtocolInfo {
            row: 16,
            col: 10,
            start_grid: 5,
            grids: phashmap![
                (
                    5,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2686,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000306,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::Begin,
                        use_perform: false,
                    }
                ),
                (
                    14,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2658,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    15,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2658,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    16,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2658,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    24,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2686,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000208,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    26,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2686,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000209,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::All,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    22,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    23,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    27,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    28,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    32,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000310,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    38,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000319,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    42,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    43,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 4,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    44,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    45,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    46,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    47,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 8,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    48,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    54,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1010707,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::BattleNormal,
                        use_perform: false,
                    }
                ),
                (
                    56,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 2000102,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::BattleNormal,
                        use_perform: false,
                    }
                ),
                (
                    64,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    66,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    73,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::BattleElite,
                        use_perform: false,
                    }
                ),
                (
                    74,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    76,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    77,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::BattleElite,
                        use_perform: false,
                    }
                ),
                (
                    83,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000313,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    87,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000317,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    93,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    94,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 6,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    95,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    96,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 10,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    97,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    104,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000318,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::BattleElite,
                        use_perform: false,
                    }
                ),
                (
                    106,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1010609,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::BattleElite,
                        use_perform: false,
                    }
                ),
                (
                    114,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 3,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    116,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 3,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    124,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 4,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1013,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    125,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2872,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 2012,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::BattleBoss,
                        use_perform: false,
                    }
                ),
                (
                    126,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 8,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1013,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    135,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000106,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::Dialog,
                        use_perform: false,
                    }
                ),
                (
                    145,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2848,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::DialogPositive,
                        use_perform: false,
                    }
                ),
                (
                    155,
                    HollowGridProtocolInfo {
                        grid: HollowGridInfo {
                            flag: 2686,
                            link_to: 15,
                            event_graph_info: HollowEventGraphInfo {
                                config_id: 0,
                                events_info: phashmap![],
                                specials: phashmap![],
                                is_new: false,
                                finished: false,
                                list_specials: phashmap![],
                                fired_count: 0,
                                hollow_event_template_id: 1000107,
                                uid: 0,
                                is_create_by_gm: false,
                            },
                            travelled_count: 0,
                            node_state: NodeState::All,
                            node_visible: NodeVisible::VisibleAtGridAround,
                        },
                        event_type: HollowEventType::BattleEnd,
                        use_perform: false,
                    }
                )
            ],
            chessboard_id: 199003501,
        });
     }
}